"""Commit Policy Checker.

Set of simple functions for dist-git policy expression evaluation as presented
at http://pkgs.devel.redhat.com/rules.html.

It uses RPN (Reverse Polish Notation) aka postfix notation for evaluation and
performs some basic syntax checks, which should be enough.
"""
from enum import IntEnum
from enum import auto
import typing


class TokenType(IntEnum):
    """Operators for Token processing."""

    # Operators
    OR = auto()
    AND = auto()
    EQ = auto()

    LPAR = auto()
    RPAR = auto()
    VAR = auto()
    LIT = auto()
    EXP = auto()


class Token:
    """Lexical token representation used for RPN evaluation."""

    # pylint: disable=too-few-public-methods

    def __init__(self, token_type: TokenType, name: str, pos: int = 0):
        """Initialize Token instance."""
        self.type = token_type
        self.name = name
        self.value: typing.Any = ''
        self.offset = -1

        if token_type is TokenType.LIT:
            self.value = name

        if pos:
            self.offset = pos - len(name)


def tokenize(
    policy: str
) -> list[Token]:
    """Parse input dist-git policy string and return list of tokens."""
    # pylint: disable=too-many-branches
    tokens: list[Token] = []
    lexeme: str = ''
    pos: int = 0
    separators: bool = True

    for pol_element in policy:
        pos += 1

        if separators and pol_element in [' ', '(', ')', '\n']:
            token_type = None

            if lexeme == 'or':
                token_type = TokenType.OR
            elif lexeme == 'and':
                token_type = TokenType.AND
            elif lexeme == '==':
                token_type = TokenType.EQ
            elif lexeme:
                if tokens and tokens[-1].type is TokenType.EQ:
                    token_type = TokenType.LIT
                    if lexeme == 'Approved' and pol_element == ' ':
                        lexeme += ' '
                        continue
                else:
                    token_type = TokenType.VAR

            if token_type:
                tokens.append(Token(token_type, lexeme, pos - 1))

            lexeme = pol_element

            if lexeme == '(':
                tokens.append(Token(TokenType.LPAR, lexeme, pos))
            elif lexeme == ')':
                tokens.append(Token(TokenType.RPAR, lexeme, pos))

            lexeme = ''

            continue

        lexeme += pol_element

        if lexeme == 'jiraField':
            separators = False
            continue

        if not separators and pol_element == ')':
            separators = True

    if lexeme:
        if tokens and tokens[-1].type == TokenType.EQ:
            token_type = TokenType.LIT
        else:
            token_type = TokenType.VAR
        tokens.append(Token(token_type, lexeme, pos))

    return tokens


def rpn(
    tokens: list[Token]
) -> list[Token]:
    """Return list of tokens in RPN.

    Input is list of tokens as returned by the tokenize function.
    """
    # pylint: disable=too-many-branches
    rpn_list: list[Token] = []
    stack: list[Token] = []
    ops = [TokenType.OR, TokenType.AND, TokenType.EQ]
    token = None

    for token in tokens:
        if token.type in ops:
            while stack:
                if stack[-1].type not in ops:
                    break
                if stack[-1].type < token.type:
                    break
                if stack[-1].type is TokenType.EQ:
                    if len(rpn_list) < 2:
                        raise Exception(
                            f'Comparison requires two operands\n{dump_tokens([stack[-1]])}')
                    if rpn_list[-1].type is not TokenType.LIT:
                        raise Exception(
                            f'Second "==" operand has to be LIT\n{dump_tokens([stack[-1]])}')
                    if rpn_list[-2].type is not TokenType.VAR:
                        raise Exception(
                            f'First "==" operand has to be VAR\n{dump_tokens([stack[-1]])}')
                rpn_list.append(stack.pop())

            stack.append(token)

        elif token.type is TokenType.LPAR:
            stack.append(token)

        elif token.type is TokenType.RPAR:
            while stack and stack[-1].type is not TokenType.LPAR:
                rpn_list.append(stack.pop())

            if not stack:
                raise Exception(f'Missing left bracket for token\n{dump_tokens([token])}')

            stack.pop()

        else:
            rpn_list.append(token)

    while stack:
        if stack[-1].type not in ops:
            raise Exception(f'Expecting EQ,OR,AND token\n{dump_tokens([stack[-1]])}')
        rpn_list.append(stack.pop())

    return rpn_list


def evaluate(
    tokens: list[Token],
    failed: list[str] | None = None,
    steps: list[Token] | None = None,
    is_jira: bool = False
) -> bool:
    """Evaluate expression represented by tokens in RPN as returned by the rpn function.

    If failed list param is set, it will contain failed bugzilla expressions. If steps list
    param is set, it will contain intermediate expressions used during evaluation. This is
    for debug purposes. Return True or False as a result of expression evaluation.
    """
    # pylint: disable=too-many-branches,too-many-statements
    stack = []
    bz_fields = ('release', 'zstream', 'blocker', 'exception',
                 'internal_target_release', 'zstream_target_release')

    for token in tokens:
        if token.type in [TokenType.LPAR, TokenType.RPAR]:
            raise Exception(f'Invalid token type\n{dump_tokens([token])}')

        if token.type in [TokenType.LIT, TokenType.VAR]:
            stack.append(token)
            continue

        if len(stack) < 2:
            raise Exception(f'Missing operand(s) for token\n{dump_tokens([token])}')

        # first operand on the stack has to be literal or expression
        op1 = stack.pop()
        if op1.type not in [TokenType.LIT, TokenType.EXP]:
            raise Exception(f'Expecting LIT,EXP token\n{dump_tokens([op1])}')

        # second operand on the stack has to be variable or expression
        op2 = stack.pop()
        if op2.type not in [TokenType.VAR, TokenType.EXP]:
            raise Exception(f'Expecting VAR,EXP token\n{dump_tokens([op2])}')

        # token with intermediate expression result
        res = Token(TokenType.EXP,
                    f'({op1.name}[{op1.value}] {token.name} {op2.name}[{op2.value}])')
        res.value = False
        res.offset = token.offset

        # do the evaluation
        if token.type is TokenType.OR:
            if op1.value or op2.value:
                res.value = True
        elif token.type is TokenType.AND:
            if op1.value and op2.value:
                res.value = True
        elif token.type is TokenType.EQ:
            if op1.value == op2.value:
                res.value = True

        stack.append(res)
        if steps is not None:
            steps.append(res)

        # store failed '==' expressions for bugzilla
        while True:
            if failed is None:
                break
            if res.value:
                break
            if op2.type is not TokenType.VAR:
                break
            # don't report failed jira expressions
            if op2.name.startswith('jira') and not is_jira:
                break
            if is_jira and op2.name in bz_fields:
                break
            if token.type is not TokenType.EQ:
                break
            failed.append(f'{op2.name}="{op2.value}" [need: "{op1.value}"]')
            break

    if len(stack) != 1:
        raise Exception(f'Unexpected stack content\n{dump_tokens(stack)}')

    res = stack.pop()
    return bool(res.value)


def errexp(
    tokens: list[Token],
    is_jira: bool = False
) -> str:
    """Return string representing the original expression extented by variable values.

    This should hopefully provide a user friendly way how to review failed expression and find out
    what conditions are not met. For better readability, by default, jira related variables are
    omitted. This can be changed by setting "is_jira" to True. The list of input tokens has
    to be the same as passed to the evaluate function. Meaning tokens in RPN and with variables
    values set.
    """
    # pylint: disable=too-many-branches,too-many-statements
    stack = []
    jira_exp = True

    for token in tokens:
        if token.type is TokenType.LIT:
            val = f'"{token.value}"'
            stack.append((not jira_exp, TokenType.LIT, val))
            continue

        if token.type is TokenType.VAR:
            val = f'{token.name} "{token.value}"'
            if token.name.startswith('jira'):
                stack.append((jira_exp, TokenType.VAR, val))
            else:
                stack.append((not jira_exp, TokenType.VAR, val))
            continue

        exp1_is_jira, exp1_type, exp1 = stack.pop()
        exp2_is_jira, exp2_type, exp2 = stack.pop()

        # If requested, by is_jira == False, remove jira
        # operator from "or" expression, but only if the other
        # operator is not jira related.
        if token.type is TokenType.OR:
            val1 = val2 = ""
            if exp1_type not in [TokenType.EQ, TokenType.OR]:
                val1 = f'({exp1})'
            else:
                val1 = exp1

            if exp2_type not in [TokenType.EQ, TokenType.OR]:
                val2 = f'({exp2})'
            else:
                val2 = exp2

            val = f'{val2} or {val1}'

            if exp1_is_jira and exp2_is_jira:
                stack.append((jira_exp, TokenType.OR, val))

            elif not exp1_is_jira and not exp2_is_jira:
                stack.append((not jira_exp, TokenType.OR, val))

            elif exp1_is_jira and not is_jira:
                stack.append((not jira_exp, exp2_type, exp2))

            elif exp2_is_jira and not is_jira:
                stack.append((not jira_exp, exp1_type, exp1))

            elif not exp1_is_jira and is_jira:
                stack.append((not jira_exp, exp2_type, exp2))

            elif not exp2_is_jira and is_jira:
                stack.append((not jira_exp, exp1_type, exp1))

            else:
                stack.append((jira_exp, TokenType.OR, val))

        elif token.type is TokenType.AND:
            val1 = val2 = ""
            if exp1_type not in [TokenType.EQ, TokenType.AND]:
                val1 = f'({exp1})'
            else:
                val1 = exp1

            if exp2_type not in [TokenType.EQ, TokenType.AND]:
                val2 = f'({exp2})'
            else:
                val2 = exp2

            val = f'{val2} and {val1}'
            if exp1_is_jira and exp2_is_jira:
                stack.append((jira_exp, TokenType.AND, val))
            else:
                stack.append((not jira_exp, TokenType.AND, val))

        elif token.type is TokenType.EQ:
            val = f'{exp2} == {exp1}'
            if exp2_is_jira:
                stack.append((jira_exp, TokenType.EQ, val))
            else:
                stack.append((not jira_exp, TokenType.EQ, val))

    _, _, val = stack.pop()
    return val


def dump_tokens(
    tokens: list[Token],
    header: bool = True
) -> str:
    """Return string/table representation of input token list."""
    text = ''

    if header:
        text += f'OFFSET TYPE {"VALUE":<35} NAME\n'
    for token in tokens:
        text += f'{token.offset:06} {token.type.name:4} {token.value:<35} {token.name}\n'

    return text.rstrip()
